from collections import namedtuple
import datetime

from .constants import *
from .classes import _AttributeList, _AttributeTagsSet

# Information about metadata attributes that can be set
# Each attribute type needs an Attribute namedtuple entry in ATTRIBUTES dict
# To add new entries, create an Attribute namedtuple and create an entry in
# ATTRIBUTES dict where key is short name for the attribute
# Fields in the Attribute namedtuple are:
# name: short name of the attribute -- will also be used as attribute/property
#   in the OSXMetaData class
# short_constant: the short constant name for the attribute
#   (e.g. kMDItemFinderComment)
# constant: the long name of the constant for the attribute
#   (e.g. com.apple.metadata:kMDItemCreator)
#   See https://developer.apple.com/documentation/coreservices/file_metadata/mditem?language=objc
#   for reference on common metadata attributes
# type_: expected type for the attribute, e.g. if Apple says it's a CFString, it'll be python str
#   CFNumber = python float, etc.
#   (called type_ so pylint doesn't complain about misplaced type identifier)
# list: (boolean) True if attribute is a list (e.g. a CFArray)
# as_list: (boolean) True if attribute is a single value but stored in a list
#   Note: the only attribute I've seen this on is com.apple.metadata:kMDItemDownloadedDate
# class: the attribute class to use, e.g. _AttributeList or str
# help: help text for the attribute (for use in command line interface)
# Note: also add short name to __slots__ in __init__.py OSXMetaData
# Note: also add the constant name (e.g. kMDItemDateAdded) to constants.py


Attribute = namedtuple(
    "Attribute",
    [
        "name",
        "short_constant",
        "constant",
        "type_",
        "list",
        "as_list",
        "class_",
        "help",
    ],
)

ATTRIBUTES = {
    "authors": Attribute(
        "authors",
        "kMDItemAuthors",
        kMDItemAuthors,
        str,
        True,
        False,
        _AttributeList,
        "The author, or authors, of the contents of the file. An array of strings.",
    ),
    "comment": Attribute(
        "comment",
        "kMDItemComment",
        kMDItemComment,
        str,
        False,
        False,
        str,
        "A comment related to the file. This differs from the Finder comment, "
        + "kMDItemFinderComment. A string.",
    ),
    "copyright": Attribute(
        "copyright",
        "kMDItemCopyright",
        kMDItemCopyright,
        str,
        False,
        False,
        str,
        "The copyright owner of the file contents. A string.",
    ),
    "creator": Attribute(
        "creator",
        "kMDItemCreator",
        kMDItemCreator,
        str,
        False,
        False,
        str,
        "Application used to create the document content (for example “Word”, “Pages”, "
        + "and so on). A string.",
    ),
    "description": Attribute(
        "description",
        "kMDItemDescription",
        kMDItemDescription,
        str,
        False,
        False,
        str,
        "A description of the content of the resource. The description may include an abstract, "
        + "table of contents, reference to a graphical representation of content or a "
        + "free-text account of the content. A string.",
    ),
    "downloadeddate": Attribute(
        "downloadeddate",
        "kMDItemDownloadedDate",
        kMDItemDownloadedDate,
        datetime.datetime,
        # False,
        True,
        # True,
        False,
        _AttributeList,
        "The date the item was downloaded.  A date in ISO 8601 format: e.g. "
        + "2000-01-12T12:00:00 or 2000-12-31 (ISO 8601 w/o time zone)",
    ),
    "findercomment": Attribute(
        "findercomment",
        "kMDItemFinderComment",
        kMDItemFinderComment,
        str,
        False,
        False,
        str,
        "Finder comments for this file. A string.",
    ),
    "headline": Attribute(
        "headline",
        "kMDItemHeadline",
        kMDItemHeadline,
        str,
        False,
        False,
        str,
        "A publishable entry providing a synopsis of the contents of the file. A string.",
    ),
    "keywords": Attribute(
        "keywords",
        "kMDItemKeywords",
        kMDItemKeywords,
        str,
        True,
        False,
        _AttributeList,
        "Keywords associated with this file. For example, “Birthday”, “Important”, etc. "
        + "This differs from Finder tags (_kMDItemUserTags) which are keywords/tags shown "
        + 'in the Finder and searchable in Spotlight using "tag:tag_name"'
        + "An array of strings.",
    ),
    "tags": Attribute(
        "tags",
        "_kMDItemUserTags",
        _kMDItemUserTags,
        str,
        True,
        False,
        _AttributeTagsSet,
        'Finder tags; searchable in Spotlight using "tag:tag_name".  '
        + "If you want tags/keywords visible in the Finder, use this instead of kMDItemKeywords.",
    ),
    "wherefroms": Attribute(
        "wherefroms",
        "kMDItemWhereFroms",
        kMDItemWhereFroms,
        str,
        True,
        False,
        _AttributeList,
        "Describes where the file was obtained from (e.g. URL downloaded from). "
        + "An array of strings.",
    ),
    # "test": Attribute(
    #     "test",
    #     "com.osxmetadata.test:DontTryThisAtHomeKids",
    #     datetime.datetime,
    #     True,
    #     False,
    #     _AttributeList,
    # ),
    # "test_float": Attribute(
    #     "test_float",
    #     "com.osxmetadata.test:DontTryThisAtHomeKids",
    #     float,
    #     False,
    #     False,
    #     float,
    # ),
    # "test_str": Attribute(
    #     "test_str", "com.osxmetadata.test:String", str, False, False, str
    # ),
}

# used for formatting output of --list
_SHORT_NAME_WIDTH = max([len(x) for x in ATTRIBUTES.keys()]) + 5
_LONG_NAME_WIDTH = max([len(x.constant) for x in ATTRIBUTES.values()]) + 10
_CONSTANT_WIDTH = 21 + 5  # currently longest is kMDItemDownloadedDate

# also add entries for attributes by constant and short constant
# do this after computing widths above
_temp_attributes = {}
for attribute in ATTRIBUTES.values():
    if attribute.constant not in ATTRIBUTES:
        _temp_attributes[attribute.constant] = attribute
        _temp_attributes[attribute.short_constant] = attribute
    else:
        raise ValueError(f"Duplicate attribute in ATTRIBUTES: {attribute}")
if _temp_attributes:
    ATTRIBUTES.update(_temp_attributes)

# list of all attributes for help text
# ATTRIBUTES_LIST = [
#     f"{'Short Name':{_SHORT_NAME_WIDTH}} {'Constant':{_CONSTANT_WIDTH}} Long Name"
# ]
# ATTRIBUTES_LIST.extend(
#     sorted(
#         [
#             f"{a.name:{_SHORT_NAME_WIDTH}} "
#             f"{a.constant.split(':',2)[1]:{_CONSTANT_WIDTH}} "
#             f"{a.constant}"
#             for a in [
#                 ATTRIBUTES[a]
#                 for a in set([attribute.name for attribute in ATTRIBUTES.values()])
#             ]
#         ]
#     )
# )



""" Python module to read and write various Mac OS X metadata 
    such as tags/keywords and Finder comments from files """


# this was inspired by osx-tags by "Ben S / scooby" and is published under
# the same MIT license. See: https://github.com/scooby/osx-tags

import datetime
import logging
import pathlib
import plistlib
import sys

# plistlib creates constants at runtime which causes pylint to complain
from plistlib import FMT_BINARY  # pylint: disable=E0611

import xattr

from .attributes import ATTRIBUTES, Attribute
from .classes import _AttributeList, _AttributeTagsSet
from .constants import (  # _DOWNLOAD_DATE,; _FINDER_COMMENT,; _TAGS,; _WHERE_FROM,
    _COLORIDS,
    _COLORNAMES,
    _FINDER_COMMENT_NAMES,
    _MAX_FINDERCOMMENT,
    _MAX_WHEREFROM,
    _VALID_COLORIDS,
    kMDItemAuthors,
    kMDItemComment,
    kMDItemCopyright,
    kMDItemCreator,
    kMDItemDescription,
    kMDItemDownloadedDate,
    kMDItemFinderComment,
    kMDItemHeadline,
    kMDItemKeywords,
    kMDItemUserTags,
    _kMDItemUserTags,
    kMDItemWhereFroms,
)
from .utils import (
    _debug,
    _get_logger,
    _set_debug,
    set_finder_comment,
    clear_finder_comment,
    validate_attribute_value,
)

__all__ = [
    "OSXMetaData",
    "ATTRIBUTES",
    "kMDItemAuthors",
    "kMDItemComment",
    "kMDItemCopyright",
    "kMDItemCreator",
    "kMDItemDescription",
    "kMDItemDownloadedDate",
    "kMDItemFinderComment",
    "kMDItemHeadline",
    "kMDItemKeywords",
    "kMDItemUserTags",
    "_kMDItemUserTags",
    "kMDItemWhereFroms",
]


# TODO: What to do about colors
# TODO: check what happens if OSXMetaData.__init__ called with invalid file--should result in error but saw one case where it didn't
# TODO: cleartags does not always clear colors--this is a new behavior, did Mac OS change something in implementation of colors?


class OSXMetaData:
    """Create an OSXMetaData object to access file metadata"""

    __slots__ = [
        "_fname",
        "_posix_name",
        "_attrs",
        "__init",
        "authors",
        "comment",
        "copyright",
        "creator",
        "description",
        "downloadeddate",
        "findercomment",
        "headline",
        "keywords",
        "tags",
        "wherefroms",
    ]

    def __init__(self, fname):
        """Create an OSXMetaData object to access file metadata"""
        self._fname = pathlib.Path(fname)
        self._posix_name = self._fname.resolve().as_posix()

        if not self._fname.exists():
            raise FileNotFoundError("file does not exist: ", fname)

        self._attrs = xattr.xattr(self._fname)

        # create property classes for the multi-valued attributes
        # tags get special handling due to color labels
        # ATTRIBUTES contains both long and short names, want only the short names (attribute.name)
        for name in set([attribute.name for attribute in ATTRIBUTES.values()]):
            attribute = ATTRIBUTES[name]
            if attribute.class_ not in [str, float, datetime.datetime]:
                super().__setattr__(name, attribute.class_(attribute, self._attrs))

        # Done with initialization
        self.__init = True

    @property
    def name(self):
        """ POSIX path of the file OSXMetaData is operating on """
        return self._fname.resolve().as_posix()

    def get_attribute(self, attribute_name):
        """ load attribute and return value or None if attribute was not set 
            attribute_name: name of attribute
        """

        attribute = ATTRIBUTES[attribute_name]
        logging.debug(f"get: {attribute}")

        # user tags need special processing to normalize names
        if attribute.name == "tags":
            return self.tags

        try:
            plist = plistlib.loads(self._attrs[attribute.constant])
        except KeyError:
            plist = None

        # TODO: should I check Attribute.type_ is correct?
        if attribute.as_list and isinstance(plist, list):
            return plist[0]
        else:
            return plist

    def get_attribute_str(self, attribute_name):
        """ returns a string representation of attribute value
            e.g. if attribute is a datedate.datetime object, will 
            format using datetime.isoformat()
            attribute_name: name of attribute """
        value = self.get_attribute(attribute_name)
        if type(value) == list or type(value) == set:
            if type(value[0]) == datetime.datetime:
                new_value = [v.isoformat() for v in value]
                return str(new_value)
            return str(value)
        else:
            if type(value) == datetime.datetime:
                return value.isoformat()
            return value

    def set_attribute(self, attribute_name, value):
        """ write attribute to file with value
            attribute_name: an osxmetadata Attribute name
            value: value to store in attribute """
        attribute = ATTRIBUTES[attribute_name]
        # verify type is correct
        if attribute.list and (type(value) == list or type(value) == set):
            for val in value:
                if attribute.type_ != type(val):
                    raise ValueError(
                        f"Expected type {attribute.type_} but value is type {type(val)}"
                    )
        elif not attribute.list and (type(value) == list or type(value) == set):
            raise TypeError(f"Expected single value but got list for {attribute.type_}")
        elif attribute.type_ != type(value):
            raise ValueError(
                f"Expected type {attribute.type_} but value is type {type(value)}"
            )

        if attribute.as_list and (type(value) != list and type(value) != set):
            # some attributes like kMDItemDownloadedDate are stored in a list
            # even though they only have only a single value
            value = [value]

        if attribute.name in _FINDER_COMMENT_NAMES:
            # Finder Comment needs special handling
            # code following will also set the attribute for Finder Comment
            set_finder_comment(self._posix_name, value)
        elif attribute.class_ in [_AttributeList, _AttributeTagsSet]:
            getattr(self, attribute.name).set_value(value)
        else:
            # must be a normal scalar (e.g. str, float)
            plist = plistlib.dumps(value, fmt=FMT_BINARY)
            self._attrs.set(attribute.constant, plist)

        return value

    def update_attribute(self, attribute_name, value):
        """ Update attribute with union of itself and value
            (this avoids adding duplicate values to attribute)
            attribute: an osxmetadata Attribute name
            value: value to append to attribute """
        return self.append_attribute(attribute_name, value, update=True)

    def append_attribute(self, attribute_name, value, update=False):
        """ append value to attribute
            attribute_name: an osxmetadata Attribute name
            value: value to append to attribute
            update: (bool) if True, update instead of append (e.g. avoid adding duplicates)
                    (default is False) """

        attribute = ATTRIBUTES[attribute_name]
        logging.debug(f"append_attribute: {attribute} {value}")

        # start with existing values
        new_value = self.get_attribute(attribute.name)

        # verify type is correct
        if attribute.list and (type(value) == list or type(value) == set):
            # expected a list, got a list
            for val in value:
                # check type of each element in list
                if attribute.type_ != type(val):
                    raise ValueError(
                        f"Expected type {attribute.type_} but value is type {type(val)}"
                    )
                else:
                    if new_value:
                        new_value = list(new_value)
                        if update:
                            # if update, only add values not already in the list
                            # behaves like set.update
                            for v in value:
                                if v not in new_value:
                                    new_value.append(v)
                        else:
                            # not update, add all values
                            new_value.extend(value)
                    else:
                        if update:
                            # no previous values but still need to make sure we don't have
                            # dupblicate values: convert to set & back to list
                            new_value = list(set(value))
                        else:
                            # no previous values, set new_value to whatever value is
                            new_value = value
        elif not attribute.list and (type(value) == list or type(value) == set):
            raise TypeError(f"Expected single value but got list for {attribute.type_}")
        else:
            # got a scalar, check type is correct
            if attribute.type_ != type(value):
                raise ValueError(
                    f"Expected type {attribute.type_} but value is type {type(value)}"
                )
            else:
                # not a list, could be str, float, datetime.datetime
                if update:
                    raise AttributeError(f"Cannot use update on {attribute.type_}")
                if new_value:
                    new_value += value
                else:
                    new_value = value

        if attribute.as_list:
            # some attributes like kMDItemDownloadedDate are stored in a list
            # even though they only have only a single value
            new_value = [new_value]

        try:
            if attribute.name in _FINDER_COMMENT_NAMES:
                # Finder Comment needs special handling
                # code following will also set the attribute for Finder Comment
                set_finder_comment(self._posix_name, new_value)

            plist = plistlib.dumps(new_value, fmt=FMT_BINARY)
            self._attrs.set(attribute.constant, plist)
        except Exception as e:
            # todo: should catch this or not?
            raise e

        return new_value

    def remove_attribute(self, attribute_name, value):
        """ remove a value from attribute, raise exception if attribute does not contain value
            only applies to multi-valued attributes, otherwise raises TypeError
            attribute_name: name of OSXMetaData attribute """

        attribute = ATTRIBUTES[attribute_name]

        if not attribute.list:
            raise TypeError("remove only applies to multi-valued attributes")

        values = self.get_attribute(attribute.name)
        values = list(values)
        values.remove(value)
        self.set_attribute(attribute.name, values)

    def discard_attribute(self, attribute_name, value):
        """ remove a value from attribute, unlike remove, does not raise exception
            if attribute does not contain value
            only applies to multi-valued attributes, otherwise raises TypeError
            attribute_name: name of OSXMetaData attribute """

        attribute = ATTRIBUTES[attribute_name]

        if not attribute.list:
            raise TypeError("discard only applies to multi-valued attributes")

        values = self.get_attribute(attribute.name)
        try:
            values.remove(value)
            self.set_attribute(attribute.name, values)
        except:
            pass

    def clear_attribute(self, attribute_name):
        """ clear anttribute (remove) 
            attribute_name: name of OSXMetaData attribute """

        attribute = ATTRIBUTES[attribute_name]

        try:
            if attribute.name in _FINDER_COMMENT_NAMES:
                # Finder Comment needs special handling
                # code following will also clear the attribute for Finder Comment
                clear_finder_comment(self._posix_name)
            self._attrs.remove(attribute.constant)
        except (IOError, OSError):
            pass

    def _list_attributes(self):
        """ list the attributes set on the file """
        return self._attrs.list()

    def list_metadata(self):
        """ list the Apple metadata attributes:
            e.g. those in com.apple.metadata namespace """
        # also lists com.osxmetadata.test used for debugging
        mdlist = self._attrs.list()
        mdlist = [
            md
            for md in mdlist
            if md.startswith("com.apple.metadata")
            or md.startswith("com.osxmetadata.test")
        ]
        return mdlist

    def __getattr__(self, name):
        """ if attribute name is in ATTRIBUTE dict, return the value
            otherwise raise KeyError """
        value = self.get_attribute(name)
        return value

    def __setattr__(self, name, value):
        """ if object is initialized and name is an attribute in ATTRIBUTES, 
            set the attribute to value
            if value value is None, will clear (delete) the attribute and all associated values
            if name is not a metadata attribute, assume it's a normal class attribute and pass to
            super() to handle  """
        try:
            if self.__init:
                # already initialized
                attribute = ATTRIBUTES[name]
                value = validate_attribute_value(attribute, value)
                if value is None:
                    self.clear_attribute(attribute.name)
                else:
                    self.set_attribute(attribute.name, value)
        except (KeyError, AttributeError):
            super().__setattr__(name, value)



# /usr/bin/env python3

import datetime
import json
import logging
import os
import os.path
import pathlib
import sys

import click

import osxmetadata

from ._version import __version__
from .attributes import _LONG_NAME_WIDTH, _SHORT_NAME_WIDTH, ATTRIBUTES
from .constants import _TAGS_NAMES
from .utils import validate_attribute_value

# TODO: add md5 option
# TODO: how is metadata on symlink handled?
# should symlink be resolved before gathering metadata?
# currently, symlinks are resolved before handling metadata but not sure this is the right behavior
# in 10.13.6: synlinks appear to inherit tags but not Finder Comments:
#   e.g. changes to tags in a file changes tags in the symlink but symlink can have it's own finder comment
#   Finder aliases inherit neither
# TODO: add selective restore (e.g only restore files matching command line path)
#   e.g osxmetadata -r meta.json *.pdf

# TODO: need special handling for --set color GREEN, etc.


# Click CLI object & context settings
class CLI_Obj:
    def __init__(self, debug=False, files=None):
        self.debug = debug
        if debug:
            osxmetadata._set_debug(True)

        self.files = files


class MyClickCommand(click.Command):
    """ Custom click.Command that overrides get_help() to show additional info """

    def get_help(self, ctx):
        help_text = super().get_help(ctx)
        formatter = click.HelpFormatter()

        # build help text from all the attribute names
        # get set of attribute names
        # (to eliminate the duplicate entries for short_constant and long costant)
        # then sort and get the short constant, long constant, and help text
        # passed to click.HelpFormatter.write_dl for formatting
        attr_tuples = [("Short Name", "Description")]
        attr_tuples.extend(
            (
                ATTRIBUTES[attr].name,
                f"{ATTRIBUTES[attr].short_constant}, "
                + f"{ATTRIBUTES[attr].constant}; {ATTRIBUTES[attr].help}",
            )
            for attr in sorted(
                [attr for attr in {attr.name for attr in ATTRIBUTES.values()}]
            )
        )

        formatter.write("\n\n")
        formatter.write_text(
            "Valid attributes for ATTRIBUTE: "
            + "Each attribute has a short name, a constant name, and a long constant name. "
            + "Any of these may be used for ATTRIBUTE"
        )
        formatter.write("\n")
        formatter.write_text('For example: --set findercomment "Hello world"')
        formatter.write_text('or:          --set kMDFinderComment "Hello world"')
        formatter.write_text(
            'or:          --set com.apple.metadata:kMDItemFinderComment "Hello world"'
        )
        formatter.write("\n")
        formatter.write_text(
            "Attributes that are strings can only take one value for --set; "
            + "--append will append to the existing value.  "
            + "Attributes that are arrays can be set multiple times to add to the array: "
            + "e.g. --set keywords 'foo' --set keywords 'bar' will set keywords to ['foo', 'bar']"
        )
        formatter.write("\n")

        formatter.write_dl(attr_tuples)
        help_text += formatter.getvalue()
        return help_text


CTX_SETTINGS = dict(help_option_names=["--help", "-h"])
FILES_ARGUMENT = click.argument(
    "files", metavar="FILE", nargs=-1, type=click.Path(exists=True)
)
WALK_OPTION = click.option(
    "--walk",
    "-w",
    is_flag=True,
    help="Walk directory tree, processing each file in the tree",
    default=False,
)
JSON_OPTION = click.option(
    "--json",
    "-j",
    "json_",
    is_flag=True,
    help="Print output in JSON format, for use with --list.",
    default=False,
    hidden=True,
)
DEBUG_OPTION = click.option(
    "--debug", required=False, is_flag=True, default=False, hidden=True
)
SET_OPTION = click.option(
    "--set",
    "set_",
    metavar="ATTRIBUTE VALUE",
    help="Set ATTRIBUTE to VALUE",
    nargs=2,
    multiple=True,
    required=False,
)
GET_OPTION = click.option(
    "--get",
    help="Get value of ATTRIBUTE",
    metavar="ATTRIBUTE",
    nargs=1,
    multiple=True,
    required=False,
)
LIST_OPTION = click.option(
    "--list",
    "list_",
    help="List all metadata attributes for FILE",
    is_flag=True,
    default=False,
)
CLEAROPTION = click.option(
    "--clear",
    help="Remove attribute from FILE",
    metavar="ATTRIBUTE",
    nargs=1,
    multiple=True,
    required=False,
)
APPEND_OPTION = click.option(
    "--append",
    metavar="ATTRIBUTE VALUE",
    help="Append VALUE to ATTRIBUTE",
    nargs=2,
    multiple=True,
    required=False,
)
UPDATE_OPTION = click.option(
    "--update",
    metavar="ATTRIBUTE VALUE",
    help="Update ATTRIBUTE with VALUE; for multi-valued attributes, "
    "this adds VALUE to the attribute if not already in the list",
    nargs=2,
    multiple=True,
    required=False,
)
REMOVE_OPTION = click.option(
    "--remove",
    metavar="ATTRIBUTE VALUE",
    help="Remove VALUE from ATTRIBUTE; only applies to multi-valued attributes",
    nargs=2,
    multiple=True,
    required=False,
)


@click.command(cls=MyClickCommand)
@click.version_option(__version__, "--version", "-v")
@DEBUG_OPTION
@FILES_ARGUMENT
@WALK_OPTION
@JSON_OPTION
@SET_OPTION
@LIST_OPTION
@CLEAROPTION
@APPEND_OPTION
@GET_OPTION
@REMOVE_OPTION
@UPDATE_OPTION
@click.pass_context
def cli(
    ctx, debug, files, walk, json_, set_, list_, clear, append, get, remove, update
):
    """ Read/write metadata from file(s). """

    if debug:
        logging.disable(logging.NOTSET)

    logging.debug(
        f"ctx={ctx} debug={debug} files={files} walk={walk} json={json_} "
        f"set={set_}, list={list_},clear={clear},append={append},get={get}, remove={remove}"
    )

    if not files:
        click.echo(ctx.get_help())
        ctx.exit()

    # validate values for --set, --clear, append, get, remove
    if any([set_, append, remove, clear, get]):
        attributes = (
            [a[0] for a in set_] + [a[0] for a in append] + list(clear) + list(get)
        )
        logging.debug(f"attributes = {attributes}")
        invalid_attr = False
        for attr in attributes:
            logging.debug(f"attr = {attr}")
            if attr not in ATTRIBUTES:
                click.echo(f"Invalid attribute {attr}", err=True)
                invalid_attr = True
        if invalid_attr:
            click.echo("")  # add a new line before rest of help text
            click.echo(ctx.get_help())
            ctx.exit()

    for f in files:
        if walk and os.path.isdir(f):
            for root, _, filenames in os.walk(f):
                # if verbose:
                #     print(f"Processing {root}")
                for fname in filenames:
                    fpath = pathlib.Path(f"{root}/{fname}").resolve()
                    process_file(
                        fpath, json_, set_, append, update, remove, clear, get, list_
                    )
        elif os.path.isdir(f):
            # skip directory
            logging.debug(f"skipping directory: {f}")
            continue
        else:
            fpath = pathlib.Path(f).resolve()
            process_file(fpath, json_, set_, append, update, remove, clear, get, list_)


def process_file(fpath, json_, set_, append, update, remove, clear, get, list_):
    """ process a single file to apply the options 
        options processed in this order: set, append, remove, clear, get, list
        Note: expects all attributes passed in parameters to be validated """

    logging.debug(f"process_file: {fpath}")

    md = osxmetadata.OSXMetaData(fpath)

    if set_:
        # set data
        # check attribute is valid
        attr_dict = {}
        for item in set_:
            attr, val = item
            attribute = ATTRIBUTES[attr]
            logging.debug(f"setting {attr}={val}")
            try:
                attr_dict[attribute].append(val)
            except KeyError:
                attr_dict[attribute] = [val]

        for attribute, value in attr_dict.items():
            logging.debug(f"value: {value}")
            value = validate_attribute_value(attribute, value)
            # tags get special handling
            if attribute.name in _TAGS_NAMES:
                tags = md.tags
                tags.clear()
                tags.update(*value)
            else:
                md.set_attribute(attribute.name, value)

    if append:
        # append data
        # check attribute is valid
        attr_dict = {}
        for item in append:
            attr, val = item
            attribute = ATTRIBUTES[attr]
            logging.debug(f"appending {attr}={val}")
            try:
                attr_dict[attribute].append(val)
            except KeyError:
                attr_dict[attribute] = [val]

        for attribute, value in attr_dict.items():
            value = validate_attribute_value(attribute, value)
            # tags get special handling
            if attribute.name in _TAGS_NAMES:
                tags = md.tags
                tags.update(*value)
            else:
                md.append_attribute(attribute.name, value)

    if update:
        # update data
        # check attribute is valid
        attr_dict = {}
        for item in update:
            attr, val = item
            attribute = ATTRIBUTES[attr]
            logging.debug(f"appending {attr}={val}")
            try:
                attr_dict[attribute].append(val)
            except KeyError:
                attr_dict[attribute] = [val]

        for attribute, value in attr_dict.items():
            value = validate_attribute_value(attribute, value)
            # tags get special handling
            if attribute.name in _TAGS_NAMES:
                tags = md.tags
                tags.update(*value)
            else:
                md.update_attribute(attribute.name, value)

    if remove:
        # remove value from attribute
        # actually implemented with discard so no error raised if not present
        # todo: catch errors and display help
        for attr, val in remove:
            try:
                attribute = ATTRIBUTES[attr]
                md.discard_attribute(attribute.name, val)
            except KeyError as e:
                raise e

    if clear:
        for attr in clear:
            attribute = ATTRIBUTES[attr]
            logging.debug(f"clearing {attr}")
            md.clear_attribute(attribute.name)

    if get:
        logging.debug(f"get: {get}")
        for attr in get:
            attribute = ATTRIBUTES[attr]
            logging.debug(f"getting {attr}")
            # tags get special handling
            if attribute.name in _TAGS_NAMES:
                value = md.tags
            else:
                value = md.get_attribute(attribute.name)
            click.echo(
                f"{attribute.name:{_SHORT_NAME_WIDTH}}{attribute.constant:{_LONG_NAME_WIDTH}} = {value}"
            )

    if list_:
        attribute_list = md.list_metadata()
        for attr in attribute_list:
            try:
                attribute = ATTRIBUTES[attr]
                # tags get special handling
                if attribute.name in _TAGS_NAMES:
                    # TODO: need to fix it so tags can be returned with proper formatting by get_attribute
                    value = md.tags
                else:
                    value = md.get_attribute_str(attribute.name)
                click.echo(
                    f"{attribute.name:{_SHORT_NAME_WIDTH}}{attribute.constant:{_LONG_NAME_WIDTH}} = {value}"
                )
            except KeyError:
                click.echo(
                    f"{'UNKNOWN':{_SHORT_NAME_WIDTH}}{attr:{_LONG_NAME_WIDTH}} = THIS ATTRIBUTE NOT HANDLED"
                )


def write_json_data(fp, data):
    json.dump(data, fp)
    fp.write("\n")


# def write_text_data(fp, data):
#     file = data["file"]

#     fc = data["fc"]
#     fc = fc if fc is not None else ""

#     dldate = data["dldate"]
#     dldate = dldate if dldate is not None else ""

#     desc = data["description"]
#     desc = desc if desc is not None else ""

#     where_from = data["where_from"]
#     where_from = where_from if where_from is not None else ""

#     tags = data["tags"]
#     tags = tags if len(tags) != 0 else ""

#     print(f"file: {file}", file=fp)
#     print(f"description: {desc}", file=fp)
#     print(f"tags: {tags}", file=fp)
#     print(f"Finder comment: {fc}", file=fp)
#     print(f"Download date: {dldate}", file=fp)
#     print(f"Where from: {where_from}", file=fp)
#     print("\n", file=fp)


# def restore_from_json(json_file, quiet=False):
#     fp = None
#     try:
#         fp = open(json_file, mode="r")
#     except:
#         print(f"Error opening file {json_file} for reading")
#         sys.exit(2)

#     for line in fp:
#         data = json.loads(line)
#         if not quiet:
#             print(f"Restoring metadata for {data['file']}")
#         set_metadata(data, quiet)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



import datetime
import logging
import os

from . import _applescript

_DEBUG = False

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(levelname)s - %(filename)s - %(lineno)d - %(message)s",
)

if not _DEBUG:
    logging.disable(logging.DEBUG)


def _get_logger():
    """Used only for testing
    
    Returns:
        logging.Logger object -- logging.Logger object for osxmetadata
    """
    return logging.Logger(__name__)


def _set_debug(debug):
    """ Enable or disable debug logging """
    global _DEBUG
    _DEBUG = debug
    if debug:
        logging.disable(logging.NOTSET)
    else:
        logging.disable(logging.DEBUG)


def _debug():
    """ returns True if debugging turned on (via _set_debug), otherwise, false """
    return _DEBUG


_scpt_set_finder_comment = _applescript.AppleScript(
    """
            on run {fname, fc}
	            set theFile to fname
	            set theComment to fc
	            tell application "Finder" to set comment of (POSIX file theFile as alias) to theComment
            end run
            """
)

_scpt_clear_finder_comment = _applescript.AppleScript(
    """
            on run {fname}
	            set theFile to fname
	            set theComment to missing value
	            tell application "Finder" to set comment of (POSIX file theFile as alias) to theComment
            end run
            """
)


def set_finder_comment(filename, comment):
    """ set finder comment for filename
        filename: path to file in posix format
        comment: comment string
    """
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"Could not find {filename}")

    _scpt_set_finder_comment.run(filename, comment)


def clear_finder_comment(filename):
    """ clear finder comment for filename
        filename: path to file in posix format
    """
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"Could not find {filename}")

    _scpt_clear_finder_comment.run(filename)


def validate_attribute_value(attribute, value):
    """ validate that value is compatible with attribute.type_ 
        and convert value to correct type
        returns value as type attribute.type_ 
        value may be a single value or a list depending on what attribute expects 
        if value contains None, returns None """

    logging.debug(
        f"validate_attribute_value: attribute: {attribute}, value: {value}, type: {type(value)}"
    )

    # check to see if we got None
    try:
        if None in value:
            return None
    except TypeError:
        if value is None:
            return None

    try:
        if isinstance(value, str):
            value = [value]
        else:
            iter(value)
    except TypeError:
        # ZZZ handle passing in values as _AttributeSet, et
        # ZZZ also handle string which can iterate but we don't want it to
        value = [value]

    # # check for None and convert to list if needed
    # if not isinstance(value, list):
    #     if value is None:
    #         return None
    #     value = [value]
    # elif None in value:
    #     return None

    if not attribute.list and len(value) > 1:
        # got a list but didn't expect one
        raise ValueError(
            f"{attribute.name} expects only one value but list of {len(value)} provided"
        )

    new_values = []
    for val in value:
        new_val = None
        if attribute.type_ == str:
            new_val = str(val)
        elif attribute.type_ == float:
            try:
                new_val = float(val)
            except:
                raise TypeError(
                    f"{val} cannot be converted to expected type {attribute.type_}"
                )
        elif attribute.type_ == datetime.datetime:
            if not isinstance(val, datetime.datetime):
                # if not already a datetime.datetime, try to convert it
                try:
                    new_val = datetime.datetime.fromisoformat(val)
                except:
                    raise TypeError(
                        f"{val} cannot be converted to expected type {attribute.type_}"
                    )
            else:
                new_val = val
        else:
            raise TypeError(f"Unknown type: {type(val)}")
        new_values.append(new_val)

    logging.debug(f"new_value = {new_values}")
    if attribute.list:
        return new_values
    else:
        return new_values[0]



""" Classes for metadata attribute types """

import collections.abc
import datetime
import plistlib
from plistlib import FMT_BINARY  # pylint: disable=E0611

from .constants import _COLORNAMES, _VALID_COLORIDS


class _AttributeList(collections.abc.MutableSequence):
    """ represents a multi-valued OSXMetaData attribute list """

    def __init__(self, attribute, xattr_):
        """ initialize object
            attribute: an OSXMetaData Attributes namedtuple 
            xattr_: an instance of xattr.xattr """
        self._attribute = attribute
        self._attrs = xattr_
        self._constant = attribute.constant

        self.data = []
        self._values = []
        self._load_data()

    def set_value(self, value):
        self.data = value
        self._write_data()

    def _load_data(self):
        self._values = []
        try:
            # load the binary plist value
            self._values = plistlib.loads(self._attrs[self._constant])
            if self._values:
                try:
                    self.data = list(self._values)
                except TypeError:
                    self.data = set([self._values])
            else:
                self.data = []
        except KeyError:
            self.data = []

    def __delitem__(self, index):
        self.data.__delitem__(index)
        self._write_data()

    def __getitem__(self, index):
        self._load_data()
        return self.data.__getitem__(index)

    def __len__(self):
        return self.data.__len__()

    def __setitem__(self, index, value):
        self.data.__setitem__(index, value)
        self._write_data()
        self._load_data()

    def insert(self, index, value):
        self.data.insert(index, value)
        self._write_data()

    def _write_data(self):
        # Overwrites the existing attribute values with the iterable of values provided.
        plist = plistlib.dumps(self.data, fmt=FMT_BINARY)
        self._attrs.set(self._constant, plist)

    def __repr__(self):
        # return f"{type(self).__name__}({super().__repr__()})"
        return repr(self.data)

    def __eq__(self, other):
        return self.data == other


class _AttributeSet:
    """ represents a multi-valued OSXMetaData attribute set """

    def __init__(self, attribute, xattr_):
        """ initialize object
            attribute: an OSXMetaData Attributes namedtuple 
            xattr_: an instance of xattr.xattr """
        self._attribute = attribute
        self._attrs = xattr_
        self._constant = attribute.constant

        # initialize
        self.data = set()
        self._load_data()

    def set_value(self, values):
        """ set value to values """
        self.data = set(map(self._normalize, values))
        self._write_data()

    def add(self, value):
        """ add a value"""
        self._load_data()
        values = set(map(self._normalize, self.data))
        self.data.add(self._normalize(value))
        self._write_data()

    def update(self, *values):
        """ update data adding any new values in *values """
        self._load_data()
        old_values = set(map(self._normalize, self.data))
        new_values = old_values.union(set(map(self._normalize, values)))
        self.data = new_values
        self._write_data()

    def clear(self):
        """ clear attribute (removes all values) """
        try:
            self._attrs.remove(self._constant)
        except (IOError, OSError):
            pass

    def remove(self, value):
        """ remove a value, raise exception if value does not exist in data set """
        self._load_data()
        values = set(map(self._normalize, self.data))
        values.remove(self._normalize(value))
        self.data = values
        self._write_data()

    def discard(self, value):
        """ remove a value, does not raise exception if value does not exist """
        self._load_data()
        values = set(map(self._normalize, self.data))
        values.discard(self._normalize(value))
        self.data = values
        self._write_data()

    def _load_data(self):
        self._values = []
        try:
            # load the binary plist value
            self._values = plistlib.loads(self._attrs[self._constant])
            if self._values:
                try:
                    self.data = set(self._values)
                except TypeError:
                    self.data = set([self._values])
            else:
                self.data = set()
        except KeyError:
            self.data = set()

    def _write_data(self):
        # TODO: change this to write_data and to read from self.data
        # Overwrites the existing tags with the iterable of tags provided.
        plist = plistlib.dumps(list(map(self._normalize, self.data)), fmt=FMT_BINARY)
        self._attrs.set(self._constant, plist)

    def _normalize(self, value):
        """ processes a value to normalize/transform the value if needed
            override in sublcass if desired (e.g. used _TagsSet) """
        return value

    def __iter__(self):
        self._load_data()
        for value in self.data:
            yield value

    def __len__(self):
        self._load_data()
        return len(self.data)

    def __repr__(self):
        self._load_data()
        return repr(self.data)

    def __str__(self):
        self._load_data()
        if self._attribute.type_ == datetime.datetime:
            values = [d.isoformat() for d in self.data]
        else:
            values = self.data
        return str(list(values))

    def __iadd__(self, value):
        for v in value:
            self.add(v)
        return self


class _AttributeTagsSet(_AttributeSet):
    """ represents a _kMDItemUserTag attribute set """

    def _tag_split(self, tag):
        # Extracts the color information from a Finder tag.

        parts = tag.rsplit("\n", 1)
        if len(parts) == 1:
            return parts[0], 0
        elif (
            len(parts[1]) != 1 or parts[1] not in _VALID_COLORIDS
        ):  # Not a color number
            return tag, 0
        else:
            return parts[0], int(parts[1])

    def _normalize(self, tag):
        """
        Ensures a color is set if not none.
        :param tag: a possibly non-normal tag.
        :return: A colorized tag.
        """
        tag, color = self._tag_split(tag)
        if tag.title() in _COLORNAMES:
            # ignore the color passed and set proper color name
            return self._tag_colored(tag.title(), _COLORNAMES[tag.title()])
        else:
            return self._tag_colored(tag, color)

    def _tag_nocolor(self, tag):
        """
        Removes the color information from a Finder tag.
        """
        return tag.rsplit("\n", 1)[0]

    def _tag_colored(self, tag, color):
        """
        Sets the color of a tag.

        Parameters:
        tag(str): a tag name
        color(int): an integer from 1 through 7

        Return:
        (str) the tag with encoded color.
        """
        return "{}\n{}".format(self._tag_nocolor(tag), color)

    def _load_data(self):
        self._tags = {}
        try:
            self._tagvalues = self._attrs[self._constant]
            # load the binary plist value
            self._tagvalues = plistlib.loads(self._tagvalues)
            for x in self._tagvalues:
                (tag, color) = self._tag_split(x)
                self._tags[tag] = color
        except KeyError:
            self._tags = None
        if self._tags:
            self.data = set(self._tags.keys())
        else:
            self.data = set()



""" constants and definitions used by osxmetadata """

# color labels
_COLORNAMES = {
    "None": 0,
    "Gray": 1,
    "Green": 2,
    "Purple": 3,
    "Blue": 4,
    "Yellow": 5,
    "Red": 6,
    "Orange": 7,
}

_COLORIDS = {
    0: "None",
    1: "Gray",
    2: "Green",
    3: "Purple",
    4: "Blue",
    5: "Yellow",
    6: "Red",
    7: "Orange",
}

_VALID_COLORIDS = "01234567"
_MAX_FINDERCOMMENT = 750  # determined through trial & error with Finder
_MAX_WHEREFROM = (
    1024
)  # just picked something....todo: need to figure out what max length is

# _TAGS = "com.apple.metadata:_kMDItemUserTags"
# _FINDER_COMMENT = "com.apple.metadata:kMDItemFinderComment"
# _WHERE_FROM = "com.apple.metadata:kMDItemWhereFroms"
# _DOWNLOAD_DATE = "com.apple.metadata:kMDItemDownloadedDate"


kMDItemAuthors = "com.apple.metadata:kMDItemAuthors"
kMDItemComment = "com.apple.metadata:kMDItemComment"
kMDItemCopyright = "com.apple.metadata:kMDItemCopyright"
kMDItemCreator = "com.apple.metadata:kMDItemCreator"
kMDItemDescription = "com.apple.metadata:kMDItemDescription"
kMDItemDownloadedDate = "com.apple.metadata:kMDItemDownloadedDate"
kMDItemFinderComment = "com.apple.metadata:kMDItemFinderComment"
kMDItemHeadline = "com.apple.metadata:kMDItemHeadline"
kMDItemKeywords = "com.apple.metadata:kMDItemKeywords"
_kMDItemUserTags = "com.apple.metadata:_kMDItemUserTags"
kMDItemUserTags = "com.apple.metadata:_kMDItemUserTags"
kMDItemWhereFroms = "com.apple.metadata:kMDItemWhereFroms"


# Special handling for Finder comments
_FINDER_COMMENT_NAMES = [
    "findercomment",
    "kMDItemFinderComment",
    "com.apple.metadata:kMDItemFinderComment",
]
_TAGS_NAMES = ["tags", "com.apple.metadata:_kMDItemUserTags"]

__all__ = [
    "kMDItemAuthors",
    "kMDItemComment",
    "kMDItemCopyright",
    "kMDItemCreator",
    "kMDItemDescription",
    "kMDItemDownloadedDate",
    "kMDItemFinderComment",
    "kMDItemHeadline",
    "kMDItemKeywords",
    "kMDItemUserTags",
    "_kMDItemUserTags",
    "kMDItemWhereFroms",
]
